import functools
import numbers

from .errors import ClaripyOperationError, ClaripyTypeError, ClaripyZeroDivisionError
from .backend_object import BackendObject
from . import debug as _d


def compare_bits(f):
    @functools.wraps(f)
    def compare_guard(self, o):
        if self.bits == 0 or o.bits == 0:
            raise ClaripyTypeError("The operation is not allowed on zero-length bitvectors.")

        if self.bits != o.bits:
            raise ClaripyTypeError("bitvectors are differently-sized (%d and %d)" % (self.bits, o.bits))
        return f(self, o)

    return compare_guard


def compare_bits_0_length(f):
    @functools.wraps(f)
    def compare_guard(self, o):
        if self.bits != o.bits:
            raise ClaripyTypeError("bitvectors are differently-sized (%d and %d)" % (self.bits, o.bits))
        return f(self, o)

    return compare_guard


def normalize_types(f):
    @functools.wraps(f)
    def normalize_helper(self, o):
        if _d._DEBUG:
            if hasattr(o, "__module__") and o.__module__ == "z3":
                raise ValueError("this should no longer happen")
        if isinstance(o, numbers.Number):
            o = BVV(o, self.bits)
        if isinstance(self, numbers.Number):
            self = BVV(self, self.bits)

        if not isinstance(self, BVV) or not isinstance(o, BVV):
            return NotImplemented
        return f(self, o)

    return normalize_helper


class BVV(BackendObject):
    __slots__ = ["bits", "_value", "mod"]

    def __init__(self, value, bits):
        if _d._DEBUG:
            if bits < 0 or not isinstance(bits, numbers.Number) or not isinstance(value, numbers.Number):
                raise ClaripyOperationError("BVV needs a non-negative length and an int value")

            if bits == 0 and value not in (0, "", None):
                raise ClaripyOperationError("Zero-length BVVs cannot have a meaningful value.")

        self.bits = bits
        self._value = 0
        self.mod = 1 << bits
        self.value = value

    def __hash__(self):
        return hash((str(self.value), self.bits))

    def __getstate__(self):
        return (self.bits, self.value)

    def __setstate__(self, s):
        self.bits = s[0]
        self.mod = 1 << self.bits
        self.value = s[1]

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v):
        self._value = v & (self.mod - 1)

    @property
    def signed(self):
        return self._value if self._value < self.mod // 2 else self._value % (self.mod // 2) - (self.mod // 2)

    @signed.setter
    def signed(self, v):
        self._value = v % -self.mod

    #
    # Arithmetic stuff
    #

    @normalize_types
    @compare_bits
    def __add__(self, o):
        return BVV(self.value + o.value, self.bits)

    @normalize_types
    @compare_bits
    def __sub__(self, o):
        return BVV(self.value - o.value, self.bits)

    @normalize_types
    @compare_bits
    def __mul__(self, o):
        return BVV(self.value * o.value, self.bits)

    @normalize_types
    @compare_bits
    def __mod__(self, o):
        if o.value == 0:
            raise ClaripyZeroDivisionError()
        return BVV(self.value % o.value, self.bits)

    @normalize_types
    @compare_bits
    def __floordiv__(self, o):
        if o.value == 0:
            raise ClaripyZeroDivisionError()
        return BVV(self.value // o.value, self.bits)

    def __truediv__(self, other):
        return self // other  # decline to implicitly have anything to do with floats

    #
    # Reverse arithmetic stuff
    #

    @normalize_types
    @compare_bits
    def __radd__(self, o):
        return BVV(self.value + o.value, self.bits)

    @normalize_types
    @compare_bits
    def __rsub__(self, o):
        return BVV(o.value - self.value, self.bits)

    @normalize_types
    @compare_bits
    def __rmul__(self, o):
        return BVV(self.value * o.value, self.bits)

    @normalize_types
    @compare_bits
    def __rmod__(self, o):
        if self.value == 0:
            raise ClaripyZeroDivisionError()
        return BVV(o.value % self.value, self.bits)

    @normalize_types
    @compare_bits
    def __rfloordiv__(self, o):
        if self.value == 0:
            raise ClaripyZeroDivisionError()
        return BVV(o.value // self.value, self.bits)

    def __rtruediv__(self, o):
        return self.__rfloordiv__(o)

    #
    # Bit operations
    #

    @normalize_types
    @compare_bits
    def __and__(self, o):
        return BVV(self.value & o.value, self.bits)

    @normalize_types
    @compare_bits
    def __or__(self, o):
        return BVV(self.value | o.value, self.bits)

    @normalize_types
    @compare_bits
    def __xor__(self, o):
        return BVV(self.value ^ o.value, self.bits)

    @normalize_types
    @compare_bits
    def __lshift__(self, o):
        return BVV(self.value << o.value, self.bits)

    @normalize_types
    @compare_bits
    def __rshift__(self, o):
        # arithmetic shift uses the signed version
        return BVV(self.signed >> o.value, self.bits)

    def __invert__(self):
        return BVV(self.value ^ self.mod - 1, self.bits)

    def __neg__(self):
        return BVV((-self.value) % self.mod, self.bits)

    #
    # Reverse bit operations
    #

    @normalize_types
    @compare_bits
    def __rand__(self, o):
        return BVV(self.value & o.value, self.bits)

    @normalize_types
    @compare_bits
    def __ror__(self, o):
        return BVV(self.value | o.value, self.bits)

    @normalize_types
    @compare_bits
    def __rxor__(self, o):
        return BVV(self.value ^ o.value, self.bits)

    @normalize_types
    @compare_bits
    def __rlshift__(self, o):
        return BVV(o.value << self.signed, self.bits)

    @normalize_types
    @compare_bits
    def __rrshift__(self, o):
        return BVV(o.signed >> self.signed, self.bits)

    #
    # Boolean stuff
    #

    @normalize_types
    @compare_bits_0_length
    def __eq__(self, o):
        return self.value == o.value

    @normalize_types
    @compare_bits_0_length
    def __ne__(self, o):
        return self.value != o.value

    @normalize_types
    @compare_bits
    def __lt__(self, o):
        return self.value < o.value

    @normalize_types
    @compare_bits
    def __gt__(self, o):
        return self.value > o.value

    @normalize_types
    @compare_bits
    def __le__(self, o):
        return self.value <= o.value

    @normalize_types
    @compare_bits
    def __ge__(self, o):
        return self.value >= o.value

    #
    # Conversions
    #

    def size(self):
        return self.bits

    def __repr__(self):
        return "BVV(0x%x, %d)" % (self.value, self.bits)


#
# External stuff
#


def BitVecVal(value, bits):
    return BVV(value, bits)


def ZeroExt(num, o):
    return BVV(o.value, o.bits + num)


def SignExt(num, o):
    return BVV(o.signed, o.bits + num)


def Extract(f, t, o):
    return BVV((o.value >> t) & ((1 << (f + 2 - t)) - 1), f + 1 - t)


def Concat(*args):
    total_bits = 0
    total_value = 0

    for o in args:
        total_value = (total_value << o.bits) | o.value
        total_bits += o.bits
    return BVV(total_value, total_bits)


def RotateRight(self, bits):
    bits_smaller = bits % self.size()
    return LShR(self, bits_smaller) | (self << (self.size() - bits_smaller))


def RotateLeft(self, bits):
    bits_smaller = bits % self.size()
    return (self << bits_smaller) | (LShR(self, (self.size() - bits_smaller)))


def Reverse(a):
    size = a.size()
    if size == 8:
        return a
    elif size % 8 != 0:
        raise ClaripyOperationError("can't reverse non-byte sized bitvectors")
    else:
        value = a.value
        out = 0
        if size == 64:
            out = _reverse_64(value)
        elif size == 32:
            out = _reverse_32(value)
        elif size == 16:
            out = _reverse_16(value)
        else:
            for i in range(0, size, 8):
                out |= ((value & (0xFF << i)) >> i) << (size - 8 - i)
        return BVV(out, size)

        # the RIGHT way to do it:
        # return BVV(int(("%x" % a.value).rjust(size/4, '0').decode('hex')[::-1].encode('hex'), 16), size)


def _reverse_16(v):
    return ((v & 0xFF) << 8) | ((v & 0xFF00) >> 8)


def _reverse_32(v):
    return ((v & 0xFF) << 24) | ((v & 0xFF00) << 8) | ((v & 0xFF0000) >> 8) | ((v & 0xFF000000) >> 24)


def _reverse_64(v):
    return (
        ((v & 0xFF) << 56)
        | ((v & 0xFF00) << 40)
        | ((v & 0xFF0000) << 24)
        | ((v & 0xFF000000) << 8)
        | ((v & 0xFF00000000) >> 8)
        | ((v & 0xFF0000000000) >> 24)
        | ((v & 0xFF000000000000) >> 40)
        | ((v & 0xFF00000000000000) >> 56)
    )


@normalize_types
@compare_bits
def ULT(self, o):
    return self.value < o.value


@normalize_types
@compare_bits
def UGT(self, o):
    return self.value > o.value


@normalize_types
@compare_bits
def ULE(self, o):
    return self.value <= o.value


@normalize_types
@compare_bits
def UGE(self, o):
    return self.value >= o.value


@normalize_types
@compare_bits
def SLT(self, o):
    return self.signed < o.signed


@normalize_types
@compare_bits
def SGT(self, o):
    return self.signed > o.signed


@normalize_types
@compare_bits
def SLE(self, o):
    return self.signed <= o.signed


@normalize_types
@compare_bits
def SGE(self, o):
    return self.signed >= o.signed


@normalize_types
@compare_bits
def SMod(self, o):
    # compute the remainder like the % operator in C
    a = self.signed
    b = o.signed
    if b == 0:
        raise ClaripyZeroDivisionError()
    division_result = a // b if a * b > 0 else (a + (-a % b)) // b
    val = a - division_result * b
    return BVV(val, self.bits)


@normalize_types
@compare_bits
def SDiv(self, o):
    # compute the round towards 0 division
    a = self.signed
    b = o.signed
    if b == 0:
        raise ClaripyZeroDivisionError()
    val = a // b if a * b > 0 else (a + (-a % b)) // b
    return BVV(val, self.bits)


#
# Pure boolean stuff
#


def BoolV(b):
    return b


def And(*args):
    return all(args)


def Or(*args):
    return any(args)


def Not(b):
    return not b


@normalize_types
def normalizer(*args):
    return args


def If(c, t, f):
    t, f = normalizer(t, f)  # pylint:disable=unbalanced-tuple-unpacking
    if c:
        return t
    else:
        return f


@normalize_types
@compare_bits
def LShR(a, b):
    return BVV(a.value >> b.value, a.bits)



# pylint:disable=duplicate-value,missing-class-docstring,wrong-import-position
import logging
import numbers
import operator
from functools import reduce

from . import BackendError, Backend

l = logging.getLogger("claripy.backends.backend_concrete")


class BackendConcrete(Backend):
    __slots__ = ()

    def __init__(self):
        Backend.__init__(self)
        self._make_raw_ops(set(backend_operations) - {"If"}, op_module=bv)
        self._make_raw_ops(backend_strings_operations, op_module=strings)
        self._make_raw_ops(backend_fp_operations, op_module=fp)
        self._op_raw["If"] = self._If
        self._op_raw["BVV"] = self.BVV
        self._op_raw["StringV"] = self.StringV
        self._op_raw["FPV"] = self.FPV

        # reduceable
        self._op_raw["__add__"] = self._op_add
        self._op_raw["__sub__"] = self._op_sub
        self._op_raw["__mul__"] = self._op_mul
        self._op_raw["__or__"] = self._op_or
        self._op_raw["__xor__"] = self._op_xor
        self._op_raw["__and__"] = self._op_and

        # unary
        self._op_raw["__invert__"] = self._op_not
        self._op_raw["__neg__"] = self._op_neg
        self._op_raw["fpSqrt"] = self._op_fpSqrt

        # boolean ops
        self._op_raw["And"] = self._op_and
        self._op_raw["Or"] = self._op_or
        self._op_raw["Xor"] = self._op_xor
        self._op_raw["Not"] = self._op_boolnot

        self._cache_objects = False

    @staticmethod
    def BVV(value, size):
        if value is None:
            raise BackendError("can't handle empty BVVs")
        return bv.BVV(value, size)

    @staticmethod
    def StringV(value, size):  # pylint: disable=unused-argument
        if not value:
            raise BackendError("can't handle empty Strings")
        return strings.StringV(value)

    @staticmethod
    def FPV(op, sort):
        return fp.FPV(op, sort)

    @staticmethod
    def _op_add(*args):
        return reduce(operator.__add__, args)

    @staticmethod
    def _op_sub(*args):
        return reduce(operator.__sub__, args)

    @staticmethod
    def _op_mul(*args):
        return reduce(operator.__mul__, args)

    @staticmethod
    def _op_or(*args):
        return reduce(operator.__or__, args)

    @staticmethod
    def _op_xor(*args):
        return reduce(operator.__xor__, args)

    @staticmethod
    def _op_and(*args):
        return reduce(operator.__and__, args)

    @staticmethod
    def _op_not(arg):
        return ~arg

    @staticmethod
    def _op_neg(arg):
        return -arg

    @staticmethod
    def _op_boolnot(arg):
        return not arg

    @staticmethod
    def _op_fpSqrt(rm, a):  # pylint:disable=unused-argument
        return a.fpSqrt()

    def convert(self, expr):
        """
        Override Backend.convert() to add fast paths for BVVs and BoolVs.
        """
        if type(expr) is BV:
            if expr.op == "BVV":
                cached_obj = self._object_cache.get(expr._cache_key, None)
                if cached_obj is None:
                    cached_obj = self.BVV(*expr.args)
                    self._object_cache[expr._cache_key] = cached_obj
                return cached_obj
        if type(expr) is Bool and expr.op == "BoolV":
            return expr.args[0]
        return super().convert(expr)

    def _If(self, b, t, f):  # pylint:disable=no-self-use,unused-argument
        if not isinstance(b, bool):
            raise BackendError("BackendConcrete can't handle non-bool condition in If.")
        return t if b else f

    def _name(self, e):  # pylint:disable=unused-argument,no-self-use
        return None

    def _identical(self, a, b):
        if type(a) is bv.BVV and type(b) is bv.BVV and a.size() != b.size():
            return False
        else:
            return a == b

    def _convert(self, a):
        if type(a) in {int, str, bytes}:
            return a
        if isinstance(a, (numbers.Number, bv.BVV, fp.FPV, fp.RM, fp.FSort, strings.StringV)):
            return a
        raise BackendError("can't handle AST of type %s" % type(a))

    def _simplify(self, e):
        return e

    def _abstract(self, e):  # pylint:disable=no-self-use
        if isinstance(e, bv.BVV):
            return BVV(e.value, e.size())
        elif isinstance(e, bool):
            return BoolV(e)
        elif isinstance(e, fp.FPV):
            return FPV(e.value, e.sort)
        elif isinstance(e, strings.StringV):
            return StringV(e.value)
        else:
            raise BackendError(f"Couldn't abstract object of type {type(e)}")

    def _cardinality(self, b):
        # if we got here, it's a cardinality of 1
        return 1

    #
    # Evaluation functions
    #

    @staticmethod
    def _to_primitive(expr):
        if isinstance(expr, bv.BVV):
            return expr.value
        elif isinstance(expr, fp.FPV):
            return expr.value
        elif isinstance(expr, strings.StringV):
            return expr.value
        elif isinstance(expr, bool):
            return expr
        elif isinstance(expr, numbers.Number):
            return expr
        else:
            raise BackendError("idk how to turn this into a primitive")

    def _eval(self, expr, n, extra_constraints=(), solver=None, model_callback=None):
        if not all(extra_constraints):
            raise UnsatError("concrete False constraint in extra_constraints")

        return (self._to_primitive(expr),)

    def _batch_eval(self, exprs, n, extra_constraints=(), solver=None, model_callback=None):
        if not all(extra_constraints):
            raise UnsatError("concrete False constraint in extra_constraints")

        return [tuple(self._to_primitive(ex) for ex in exprs)]

    def _max(self, expr, extra_constraints=(), signed=False, solver=None, model_callback=None):
        if not all(extra_constraints):
            raise UnsatError("concrete False constraint in extra_constraints")
        return self._to_primitive(expr)

    def _min(self, expr, extra_constraints=(), signed=False, solver=None, model_callback=None):
        if not all(extra_constraints):
            raise UnsatError("concrete False constraint in extra_constraints")
        return self._to_primitive(expr)

    def _solution(self, expr, v, extra_constraints=(), solver=None, model_callback=None):
        if not all(extra_constraints):
            raise UnsatError("concrete False constraint in extra_constraints")
        return self.convert(expr) == v

    # Override Backend.is_true() for a better performance
    def is_true(self, e, extra_constraints=(), solver=None, model_callback=None):
        if e in {True, 1, 1.0}:
            return True
        if e in {False, 0, 0.0}:
            return False
        if type(e) is Base and e.op == "BoolV" and len(e.args) == 1 and e.args[0] is True:
            return True
        return super().is_true(e, extra_constraints=extra_constraints, solver=solver, model_callback=model_callback)

    # Override Backend.is_false() for a better performance
    def is_false(self, e, extra_constraints=(), solver=None, model_callback=None):
        if e in {False, 0, 0.0}:
            return True
        if e in {True, 1, 1.0}:
            return False
        if type(e) is Base and e.op == "BoolV" and len(e.args) == 1 and e.args[0] is False:
            return True
        return super().is_false(e, extra_constraints=extra_constraints, solver=solver, model_callback=model_callback)

    # pylint:disable=singleton-comparison
    def _is_true(self, e, extra_constraints=(), solver=None, model_callback=None):
        return e == True

    def _is_false(self, e, extra_constraints=(), solver=None, model_callback=None):
        return e == False

    def _has_true(self, e, extra_constraints=(), solver=None, model_callback=None):
        return e == True

    def _has_false(self, e, extra_constraints=(), solver=None, model_callback=None):
        return e == False


from ..operations import backend_operations, backend_fp_operations, backend_strings_operations
from .. import bv, fp, strings
from ..ast import Base
from ..ast.bv import BV, BVV
from ..ast.strings import StringV
from ..ast.fp import FPV
from ..ast.bool import Bool, BoolV
from ..errors import UnsatError



from typing import Union, List, Optional, overload, Any

from .bits import Bits
from .bool import Bool
from .fp import FP
from ..bv import BVV as ConcreteBVV
from ..fp import RM, FSort

CoerceBV = Union[int, BV]
Incomplete = Any

class BV(Bits):
    def chop(self, bits: int = 1) -> List[BV]: ...
    def __getitem__(self, rng: Union[int, slice]) -> BV: ...
    def get_byte(self, index: int) -> BV: ...
    def get_bytes(self, index: int, size: int) -> BV: ...
    def zero_extend(self, n: int) -> BV: ...
    def sign_extend(self, n: int) -> BV: ...
    def concat(self, *args: BV) -> BV: ...
    @staticmethod
    def _from_int(like, value: int) -> BV: ...
    @staticmethod
    def _from_Bool(like, value: Bool) -> BV: ...
    @staticmethod
    def _from_bytes(like, value: bytes) -> BV: ...
    @staticmethod
    def _from_str(like, value: str) -> BV: ...
    @staticmethod
    def _from_BVV(like, value: ConcreteBVV) -> BV: ...
    def val_to_fp(self, sort: FSort, signed: bool = True, rm: Optional[RM] = None) -> FP: ...
    def raw_to_fp(self) -> FP: ...
    def raw_to_bv(self) -> BV: ...
    def to_bv(self) -> BV: ...
    def __add__(self, other: CoerceBV) -> BV: ...
    def __radd__(self, other: CoerceBV) -> BV: ...
    def __floordiv__(self, other: CoerceBV) -> BV: ...
    def __rfloordiv__(self, other: CoerceBV) -> BV: ...
    def __truediv__(self, other: CoerceBV) -> BV: ...
    def __rtruediv__(self, other: CoerceBV) -> BV: ...
    def __mul__(self, other: CoerceBV) -> BV: ...
    def __rmul__(self, other: CoerceBV) -> BV: ...
    def __sub__(self, other: CoerceBV) -> BV: ...
    def __rsub__(self, other: CoerceBV) -> BV: ...
    def __pow__(self, other: CoerceBV) -> BV: ...
    def __rpow__(self, other: CoerceBV) -> BV: ...
    def __mod__(self, other: CoerceBV) -> BV: ...
    def __rmod__(self, other: CoerceBV) -> BV: ...
    def SDiv(self, other: CoerceBV) -> BV: ...
    def SMod(self, other: CoerceBV) -> BV: ...
    def __or__(self, other: CoerceBV) -> BV: ...
    def __ror__(self, other: CoerceBV) -> BV: ...
    def __and__(self, other: CoerceBV) -> BV: ...
    def __rand__(self, other: CoerceBV) -> BV: ...
    def __xor__(self, other: CoerceBV) -> BV: ...
    def __rxor__(self, other: CoerceBV) -> BV: ...
    def __lshift__(self, other: CoerceBV) -> BV: ...
    def __rlshift__(self, other: CoerceBV) -> BV: ...
    def __rshift__(self, other: CoerceBV) -> BV: ...
    def __rrshift__(self, other: CoerceBV) -> BV: ...
    def LShR(self, other: CoerceBV) -> BV: ...
    def __neg__(self) -> BV: ...
    def __pos__(self) -> BV: ...
    def __abs__(self) -> BV: ...
    def __invert__(self) -> BV: ...
    def __eq__(self, other: CoerceBV) -> Bool: ...
    def __ne__(self, other: CoerceBV) -> Bool: ...
    def __ge__(self, other: CoerceBV) -> Bool: ...
    def __le__(self, other: CoerceBV) -> Bool: ...
    def __gt__(self, other: CoerceBV) -> Bool: ...
    def __lt__(self, other: CoerceBV) -> Bool: ...
    def SLT(self, other: CoerceBV) -> Bool: ...
    def SLE(self, other: CoerceBV) -> Bool: ...
    def SGT(self, other: CoerceBV) -> Bool: ...
    def SGE(self, other: CoerceBV) -> Bool: ...
    def ULT(self, other: CoerceBV) -> Bool: ...
    def ULE(self, other: CoerceBV) -> Bool: ...
    def UGT(self, other: CoerceBV) -> Bool: ...
    def UGE(self, other: CoerceBV) -> Bool: ...
    def Extract(self, start: int, end: int) -> BV: ...
    def Concat(self, other: BV) -> BV: ...
    @property
    def reversed(self) -> BV: ...

    # weh
    def union(self, other: BV) -> BV: ...
    def widen(self, other: BV) -> BV: ...
    def intersection(self, other: BV) -> BV: ...
    @property
    def concrete_value(self) -> int: ...

def BVS(
    name: str,
    size: int,
    min=None,
    max=None,
    stride=None,
    uninitialized=False,  # pylint:disable=redefined-builtin
    explicit_name=None,
    discrete_set=False,
    discrete_set_max_card=None,
    **kwargs,
) -> BV: ...
@overload
def BVV(value: int, size: int = ..., **kwargs) -> BV: ...
@overload
def BVV(value: bytes, **kwargs) -> BV: ...
def SI(
    name: str | None = ...,
    bits: int = ...,
    lower_bound: int | None = ...,
    upper_bound: int | None = ...,
    stride: int | None = ...,
    to_conv: Incomplete | None = ...,
    explicit_name: bool | None = ...,
    discrete_set: bool = ...,
    discrete_set_max_card: int | None = ...,
): ...
def TSI(bits, name: str | None = ..., uninitialized: bool = ..., explicit_name: bool | None = ...): ...
def ESI(bits, **kwargs): ...
def ValueSet(
    bits,
    region: str | None = ...,
    region_base_addr: int | None = ...,
    value: int | None = ...,
    name: str | None = ...,
    val: Incomplete | None = ...,
): ...

VS = ValueSet

def DSIS(
    name: Incomplete | None = ...,
    bits: int = ...,
    lower_bound: Incomplete | None = ...,
    upper_bound: Incomplete | None = ...,
    stride: Incomplete | None = ...,
    explicit_name: Incomplete | None = ...,
    to_conv: Incomplete | None = ...,
    max_card: Incomplete | None = ...,
): ...
def ULT(a: Union[BV, int], b: Union[BV, int]) -> Bool: ...
def ULE(a: Union[BV, int], b: Union[BV, int]) -> Bool: ...
def UGT(a: Union[BV, int], b: Union[BV, int]) -> Bool: ...
def UGE(a: Union[BV, int], b: Union[BV, int]) -> Bool: ...
def SLT(a: Union[BV, int], b: Union[BV, int]) -> Bool: ...
def SLE(a: Union[BV, int], b: Union[BV, int]) -> Bool: ...
def SGT(a: Union[BV, int], b: Union[BV, int]) -> Bool: ...
def SGE(a: Union[BV, int], b: Union[BV, int]) -> Bool: ...
def SDiv(a: Union[BV, int], b: Union[BV, int]) -> BV: ...
def SMod(a: Union[BV, int], b: Union[BV, int]) -> BV: ...
def LShR(a: Union[BV, int], b: Union[BV, int]) -> BV: ...
def SignExt(a: int, b: Union[BV, int]) -> BV: ...
def ZeroExt(a: int, b: Union[BV, int]) -> BV: ...
def Extract(a: int, b: int, BV) -> BV: ...
def Concat(*a: BV) -> BV: ...
def RotateLeft(a: Union[BV, int], b: Union[BV, int]) -> BV: ...
def RotateRight(a: Union[BV, int], b: Union[BV, int]) -> BV: ...
def Reverse(a: Union[BV, int]) -> BV: ...
def union(a: Union[BV, int], b: Union[BV, int]) -> BV: ...
def widen(a: Union[BV, int], b: Union[BV, int]) -> BV: ...
def intersection(a: Union[BV, int], b: Union[BV, int]) -> BV: ...



call_depth = 0


def _log(s):
    print("|  " * call_depth + s)


def _debugged(f):
    def debugged(*args, **kwargs):
        global call_depth

        _log(f.__name__ + " " + str(args) + " " + str(kwargs))
        call_depth += 1
        try:
            r = f(*args, **kwargs)
        except Exception as e:  # pylint:disable=broad-except
            r = "EXCEPTION: %s" % str(e)
            raise
        finally:
            call_depth -= 1
            if r is not None:
                _log(r"\... " + str(r))

        if hasattr(r, "__iter__") and hasattr(r, "next"):
            return _debug_iterator(r)
        else:
            return r

    return debugged


def _debug_iterator(i):
    for v in i:
        _log("NEXT: " + str(v))
        yield v
    _log("FINISHED: " + str(i))


class DebugMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getattribute__(self, a):
        o = super().__getattribute__(a)
        if a.startswith("__"):
            return o
        elif hasattr(o, "__call__"):
            return _debugged(o)
        elif hasattr(o, "__next__"):
            return _debug_iterator(o)
        else:
            return o


def debug_decorator(o):
    if isinstance(o, type):

        class Debugged(DebugMixin, o):
            pass

        Debugged.__name__ = "Debugged" + o.__name__
        return Debugged
    elif hasattr(o, "__call__"):
        return _debugged(o)



import unittest
import claripy
from claripy.backends.backend_smtlib_solvers.z3_popen import SolverBackendZ3
import common_backend_smt_solver


class SmtLibSolverTest_Z3(common_backend_smt_solver.SmtLibSolverTestBase):
    @unittest.skip("Skip these test for now because of a problem with pysmt")
    def get_solver(self):
        backend = SolverBackendZ3(daggify=True)
        solver = claripy.SolverStrings(backend=backend)
        return solver


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(SmtLibSolverTest_Z3)
    unittest.TextTestRunner(verbosity=2).run(suite)



import logging

from claripy.vsa import StridedInterval

l = logging.getLogger("angr_tests")


def check_si_fields(si, stride, lb, ub):
    if si.stride != stride:
        return False
    if si.lower_bound != lb:
        return False
    if si.upper_bound != ub:
        return False
    return True


def test_smart_join():
    s1 = StridedInterval(bits=4, stride=3, lower_bound=9, upper_bound=12)
    s2 = StridedInterval(bits=4, stride=3, lower_bound=0, upper_bound=3)
    j = StridedInterval.pseudo_join(s1, s2)
    u = StridedInterval.least_upper_bound(s1, s2)
    assert check_si_fields(u, 3, 0, 12)
    assert check_si_fields(j, 3, 0, 12)

    s1 = StridedInterval(bits=4, stride=0, lower_bound=8, upper_bound=8)
    s2 = StridedInterval(bits=4, stride=1, lower_bound=14, upper_bound=15)
    s3 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=4)
    u = StridedInterval.least_upper_bound(s1, s2, s3)
    assert check_si_fields(u, 1, 14, 8)

    s1 = StridedInterval(bits=4, stride=3, lower_bound=2, upper_bound=8)
    s2 = StridedInterval(bits=4, stride=0, lower_bound=1, upper_bound=1)
    j = StridedInterval.pseudo_join(s1, s2)
    u = StridedInterval.least_upper_bound(s1, s2)
    assert check_si_fields(u, 3, 2, 1)
    assert check_si_fields(j, 3, 2, 1)


if __name__ == "__main__":
    test_smart_join()



#!/usr/bin/env python

import logging
import numbers

l = logging.getLogger("claripy.frontends.frontend")


class Frontend:
    def __init__(self):
        pass

    def __getstate__(self):
        return True  # need to return something so that pickle calls setstate

    def __setstate__(self, s):  # pylint:disable=unused-argument
        return

    def branch(self):
        c = self.blank_copy()
        self._copy(c)
        return c

    def blank_copy(self):
        c = self.__class__.__new__(self.__class__)
        self._blank_copy(c)
        return c

    def _blank_copy(self, c):  # pylint:disable=no-self-use,unused-argument
        return

    def _copy(self, c):  # pylint:disable=no-self-use,unused-argument
        return

    #
    # Stuff that should be implemented by subclasses
    #

    def eval_to_ast(self, e, n, extra_constraints=(), exact=None):
        """
        Evaluates expression `e`, returning a list of `n` concrete ASTs.

        :param e:                       the expression
        :param n:                       the number of ASTs to return
        :param extra_constraints:       extra constraints to consider when performing the evaluation
        :param exact:                   whether or not to perform an exact evaluation. Ignored by
                                        non-approximating backends.

        :return:                        list of concrete ASTs
        """
        return [ast.bv.BVV(v, e.size()) for v in self.eval(e, n, extra_constraints=extra_constraints, exact=exact)]

    def finalize(self):
        raise NotImplementedError("finalize() is not implemented")

    def merge(self, others, merge_conditions, common_ancestor=None):
        raise NotImplementedError("merge() is not implemented")

    def combine(self, others):
        raise NotImplementedError("combine() is not implemented")

    def split(self):
        raise NotImplementedError("split() is not implemented")

    def add(self, constraints):
        """
        Adds constraint(s) to constraints list.

        :param constraints:             constraint(s) to add

        :return:
        """
        raise NotImplementedError()

    def simplify(self):
        """
        Simplifies the stored constraints conjunction.
        """
        raise NotImplementedError()

    def check_satisfiability(self, extra_constraints=(), exact=None):
        """
        Checks the satisfiability of stored constraints conjunction.

        :param extra_constraints:       extra constraints to consider when checking satisfiability
        :param exact:                   whether or not to perform exact checking. Ignored by
                                        non-approximating backends.

        :return:                        'SAT' if the conjunction is satisfiable otherwise 'UNSAT'
        """
        raise NotImplementedError

    def satisfiable(self, extra_constraints=(), exact=None):
        """
        Checks if stored constraints conjunction is satisfiable.

        :param extra_constraints:       extra constraints to consider when checking satisfiability
        :param exact:                   whether or not to perform exact checking. Ignored by
                                        non-approximating backends.

        :return:                        True if the conjunction is satisfiable otherwise False
        """
        raise NotImplementedError()

    def eval(self, e, n, extra_constraints=(), exact=None):
        """
        Evaluates expression `e`, returning a tuple of `n` solutions.

        :param e:                       the expression
        :param n:                       the number of solutions to return
        :param extra_constraints:       extra constraints to consider when performing the evaluation
        :param exact:                   whether or not to perform an exact evaluation. Ignored by
                                        non-approximating backends.

        :return:                        tuple of python primitives representing results
        """
        raise NotImplementedError()

    def batch_eval(self, exprs, n, extra_constraints=(), exact=None):
        """
        Evaluates `exprs`, returning a list of tuples (one tuple of `n` solutions for expression).

        :param exprs:                   expressions
        :param n:                       the number of solutions to return
        :param extra_constraints:       extra constraints to consider when performing the evaluation
        :param exact:                   whether or not to perform an exact evaluation. Ignored by
                                        non-approximating backends.

        :return:                        list of tuples of python primitives representing results
        """
        raise NotImplementedError()

    def max(self, e, extra_constraints=(), signed=False, exact=None):
        """
        Evaluates `e`, returning its max possible value.

        :param e:                       the expression
        :param extra_constraints:       extra constraints to consider when performing the evaluation
        :param signed:                  whether the value should be treated as a signed integer
        :param exact:                   whether or not to perform an exact evaluation. Ignored by
                                        non-approximating backends.

        :return:                        max possible value
        """
        raise NotImplementedError()

    def min(self, e, extra_constraints=(), signed=False, exact=None):
        """
        Evaluates `e`, returning its min possible value.

        :param e:                       the expression
        :param extra_constraints:       extra constraints to consider when performing the evaluation
        :param signed:                  whether the value should be treated as a signed integer
        :param exact:                   whether or not to perform an exact evaluation. Ignored by
                                        non-approximating backends.

        :return:                        min possible value
        """
        raise NotImplementedError()

    def solution(self, e, v, extra_constraints=(), exact=None):
        """
        Checks if `v` is a possible solution to `e`.

        :param e:                       the expression
        :param v:                       the value
        :param extra_constraints:       extra constraints to consider when performing the evaluation
        :param exact:                   whether or not to perform an exact evaluation. Ignored by
                                        non-approximating backends.

        :return:                        True if it is a possible solution otherwise False
        """
        raise NotImplementedError()

    def is_true(self, e, extra_constraints=(), exact=None):
        """
        Checks if `e` can only (and TRIVIALLY) evaluate to True. If this function returns True,
        then the expression cannot ever be False, regardless of constraints or anything else.
        If the expression returns False, then the expression might STILL not ever be False; it's just
        that we can't trivially prove it. In other words, a return value of False gives you no
        information whatsoever.


        :param e:                       the expression
        :param extra_constraints:       extra constraints to consider when performing the evaluation
        :param exact:                   whether or not to perform an exact evaluation. Ignored by
                                        non-approximating backends.

        :return:                        True if it can only evaluate to True otherwise False
        """
        raise NotImplementedError()

    def is_false(self, e, extra_constraints=(), exact=None):
        """
        Checks if `e` can only (and TRIVIALLY) evaluate to False. If this function returns True,
        then the expression cannot ever be True, regardless of constraints or anything else.
        If the expression returns False, then the expression might STILL not ever be True; it's just
        that we can't trivially prove it. In other words, a return value of False gives you no
        information whatsoever.

        :param e:                       the expression
        :param extra_constraints:       extra constraints to consider when performing the evaluation
        :param exact:                   whether or not to perform an exact evaluation. Ignored by
                                        non-approximating backends.

        :return:                        True if it can only evaluate to False otherwise False
        """
        raise NotImplementedError()

    def downsize(self):  # pylint:disable=no-self-use
        pass

    #
    # Some utility functions
    #

    def _concrete_value(self, e):  # pylint:disable=no-self-use
        if isinstance(e, numbers.Number):
            return e
        else:
            return None

    _concrete_constraint = _concrete_value

    def _constraint_filter(self, c):  # pylint:disable=no-self-use
        return c

    @staticmethod
    def _split_constraints(constraints, concrete=True):
        """
        Returns independent constraints, split from this Frontend's `constraints`.
        """

        splitted = []
        for i in constraints:
            splitted.extend(i.split(["And"]))

        l.debug("... splitted of size %d", len(splitted))

        concrete_constraints = []
        variable_connections = {}
        constraint_connections = {}
        for n, s in enumerate(splitted):
            l.debug("... processing constraint with %d variables", len(s.variables))

            connected_variables = set(s.variables)
            connected_constraints = {n}

            if len(connected_variables) == 0:
                concrete_constraints.append(s)

            for v in s.variables:
                if v in variable_connections:
                    connected_variables |= variable_connections[v]
                if v in constraint_connections:
                    connected_constraints |= constraint_connections[v]

            for v in connected_variables:
                variable_connections[v] = connected_variables
                constraint_connections[v] = connected_constraints

        unique_constraint_sets = set()
        for v in variable_connections:
            unique_constraint_sets.add((frozenset(variable_connections[v]), frozenset(constraint_connections[v])))

        results = []
        for v, c_indexes in unique_constraint_sets:
            results.append((set(v), [splitted[c] for c in c_indexes]))

        if concrete and len(concrete_constraints) > 0:
            results.append(({"CONCRETE"}, concrete_constraints))

        return results


from . import ast
